(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory InnermostImpl imports Innermost Substitution TermImpl begin
      

fun NF_lhs_list :: "('f,'v)term list \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "NF_lhs_list [] _ = True"
    | "NF_lhs_list Q t = (list_all (\<lambda> st. list_all (\<lambda> q. \<not> (matches st q)) Q) (supteq_list t))"

lemma NF_lhs_list[simp]: "NF_lhs_list Q = NF_lhs (set Q)"
proof (rule ext)
  fix t
  show "NF_lhs_list Q t = NF_lhs (set Q) t"
  proof (cases Q, simp add: NF_lhs_empty)
    case (Cons q Q')
    hence "NF_lhs_list Q t = (list_all (\<lambda> st. list_all (\<lambda> q. \<not> matches st q) Q) (supteq_list t))" by simp
    also have "\<dots> = (\<forall> s. t \<unrhd> s \<longrightarrow> (\<forall> q \<in> set Q. \<not> matches s q))" by (simp add: list_all_iff)
    also have "\<dots> = (\<forall> C q \<sigma>. q \<in> set Q \<longrightarrow> t \<noteq> C\<langle>q \<cdot> \<sigma>\<rangle>)" (is "?l = ?r")
    proof
      assume ?l
      show ?r
      proof (rule ccontr)
        assume "\<not> ?r"
        then obtain C q \<sigma> where q: "q \<in> set Q" and "t = C\<langle>q \<cdot> \<sigma>\<rangle>" by auto
        with ctxt_imp_supteqp[of C "q \<cdot> \<sigma>"] `?l` show False by auto
      qed
    next
      assume ?r
      show ?l
      proof (rule ccontr)
        assume "\<not> ?l"
        then obtain s q \<sigma> where sub: "t \<unrhd> s" and "q \<in> set Q" and "s = q \<cdot> \<sigma>" by auto
        with `?r` supteqp_imp_ctxt[OF sub] show False by auto
      qed
    qed          
    also have "\<dots> = NF_lhs (set Q) t" unfolding NF_lhs_def by auto
    finally show "NF_lhs_list Q t = NF_lhs (set Q) t" .
  qed
qed

declare NF_lhs_list.simps[simp del]
(* TODO 
definition NF_Q_R_check :: "('f,'v)qtrs \<Rightarrow> bool"
where "NF_Q_R_check QR \<equiv> (\<forall> l \<in> fst ` (snd QR). (\<exists> q C \<sigma>. q \<in> fst QR \<and> l = C\<langle>q \<cdot> \<sigma>\<rangle>))"
*)

fun useful_rule_list :: "('f,'v)term list \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "useful_rule_list [] lr = True"
   |  "useful_rule_list _ (Var _, _) = True"
   |  "useful_rule_list Q (Fun f ss, _) = list_all (NF_lhs_list Q) ss"

lemma useful_rule_list[simp]: "useful_rule_list Q = useful_rule (set Q)"
proof (rule ext, simp only: split_paired_all)
  fix l r
  show "useful_rule_list Q (l,r) = useful_rule (set Q) (l,r)"
  proof (cases Q, simp add: useful_rule_empty)
    case (Cons q Q')
    thus ?thesis
    proof (cases l, simp add: left_var_useful)
      case (Fun f ss)
      with Cons have "useful_rule_list Q (l,r) = (\<forall> s \<in> set ss. NF_lhs (set Q) s)" by (simp add: list_all_iff)
      also have "\<dots> = (\<forall> u. Fun f ss \<rhd> u \<longrightarrow> NF_lhs (set Q) u)" by (simp only: NF_lhs_suptp_is_NF_lhs_args, blast)
      also have "\<dots> = useful_rule (set Q) (Fun f ss,r)" (is "?l = ?r")
      proof
        assume ?l
        show ?r
        proof (rule ccontr)
          assume "\<not> ?r"
          then obtain ll C \<sigma> where ll: "ll \<in> set Q" and "Fun f ss \<rhd> ll \<cdot> \<sigma>" unfolding useful_rule_def by auto
          with `?l` have NF: "NF_lhs (set Q) (ll \<cdot> \<sigma>)" by auto
          from NF_lhs_instance[OF NF] not_NF_lhs_lhs[OF ll]
          show False by simp
        qed
      next
        assume ?r
        show ?l
        proof (intro allI impI)
          fix u
          assume supt: "Fun f ss \<rhd> u"
          show "NF_lhs (set Q) u"
          proof (rule ccontr)
            assume "\<not> ?thesis"
            then obtain q C \<sigma> where u: "u = C\<langle>q \<cdot> \<sigma>\<rangle>" and q: "q \<in> set Q" unfolding NF_lhs_def by auto 
            from suptp_imp_nectxt[OF supt] obtain D where D: "D \<noteq> \<box>" and "Fun f ss = D\<langle>u\<rangle>" by auto
            hence id: "Fun f ss = (D \<circ>c C)\<langle>q \<cdot> \<sigma>\<rangle>" by (simp add: u)
            from D have "D \<circ>c C \<noteq> \<box>" by (cases D, simp, cases C, auto)
            hence "Fun f ss \<rhd> q \<cdot> \<sigma>" using nectxt_imp_suptp[OF _ id] by simp
            with `?r` q show False unfolding useful_rule_def by auto
          qed
        qed
      qed
      finally show ?thesis using Fun by simp
    qed
  qed
qed
declare useful_rule_list.simps[simp del]

types ('f,'v)qrules = "('f,'v)term list \<times> ('f,'v)rules"    

fun useful_rules_list :: "('f,'v)qrules \<Rightarrow> ('f,'v)rules"
where "useful_rules_list ([],R) = R"
   |  "useful_rules_list (Q,R) = filter (useful_rule_list Q) R"

lemma useful_rules_list[simp]: "set (useful_rules_list (Q,R)) = useful_rules (set Q, set R)"
proof (cases Q, simp add: useful_rules_empty)
  case (Cons q Q')
  thus ?thesis unfolding useful_rules_def by auto
qed

declare useful_rules_list.simps[simp del]

(* TODO 
definition wwf_rule :: "('f,'v)term set \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "wwf_rule Q lr \<equiv> case lr of Pair l r \<Rightarrow> useful_rule Q lr \<longrightarrow> ((\<exists> f ts. l = Fun f ts) \<and> vars_term l \<supseteq> vars_term r)"

definition wwf_qtrs :: "('f,'v)qtrs \<Rightarrow> bool"
where "wwf_qtrs QR \<equiv> \<forall> lr \<in> snd QR. wwf_rule (fst QR) lr"
*)

end
