(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RPO
imports TermOrder LexicographicExtension
begin

(* precedence, max arity, and two terms are given, result is pair (strict, non-strict) *)
function
  lpo :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool"
where
  "lpo _ _ (Var x) (Var y) = (False, x = y)" |
  "lpo pr _ (Var x) (Fun g ts) = (False, ts = [] \<and> pr g 0 = 0)" |
  "lpo pr n (Fun f ss) (Var y) = (let con = list_ex (\<lambda> s. snd (lpo pr n s (Var y))) ss in (con,con))" |
  "lpo pr n (Fun f ss) (Fun g ts) = (
    if list_ex (\<lambda> s. snd (lpo pr n s (Fun g ts))) ss 
       then (True,True)
       else (let prl = pr f (length ss); prr = pr g (length ts) in 
         if prl \<ge> prr \<and> (list_all (\<lambda> t. fst (lpo pr n (Fun f ss) t)) ts)
          then if prl > prr then (True,True) 
            else lex_ext (lpo pr n) n ss ts
          else (False,False)))"
by pat_completeness auto

(* locally add these lemmas to the simplifier *)
declare list_all_ex[simp]
(* and at the end throw them out of the simplifier again *)


termination lpo 
proof (relation "measures [\<lambda> (pr,n,s,t). size s + size t]", (auto)[4])
  fix "pr" n and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa xb xc i
  assume "i < length ss" and "i < length ts"
  then obtain s t where si: "s = ss ! i" and s: "s \<in> set ss" and ti: "t = ts ! i" and t: "t \<in> set ts" by auto
  from s have s1: "size s \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size t \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 si ti show "((pr, n, ss ! i, ts ! i), pr, n, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,n,s,t). size s + size t]"
    by auto
qed

lemma lpo_stri_imp_nstri[rule_format] :  "fst (lpo pr n s t) \<longrightarrow>
  snd (lpo pr n s t)"
using assms
proof (induct rule: lpo.induct[of "\<lambda> pr n s t. fst (lpo pr n s t) \<longrightarrow> snd (lpo pr n s t)"], (auto simp: Let_def)[3])
  case (4 "pr" n f ss g ts)
  show ?case 
    by (cases "length ss = length ts \<or> length ts \<le> n", auto simp: Let_def lex_ext_stri_imp_nstri) 
qed

lemma lpo_nstri_refl : "snd (lpo pr n s s)"
proof (induct s, simp)
  case (Fun f ss)
  have rec1: "snd (lex_ext (lpo pr n) n ss ss)"
    by (rule all_nstri_imp_lex_nstri, auto simp: Fun)
  have rec2: "\<forall> s \<in> set ss. fst (lpo pr n (Fun f ss) s)"
  proof
    fix s
    assume s: "s \<in> set ss"
    with Fun have rec: "snd (lpo pr n s s)" by auto
    with s show "fst (lpo pr n (Fun f ss) s)"
    proof (cases s)
      case (Fun g ts)
      from s rec have "list_ex (\<lambda> t. snd (lpo pr n t s)) ss" by (auto)
      with Fun show ?thesis by auto
    next
      case (Var x)
      from s rec have "list_ex (\<lambda> t. snd (lpo pr n t s)) ss" by auto
      with Var show ?thesis by auto
    qed
  qed
  from rec1 rec2 show ?case by (auto simp: Let_def)
qed

lemma supt_imp_lpo_stri : assumes "s \<rhd> t" 
  shows "fst (lpo pr n s t)"
  using assms
proof (induct)
  case (supt_arg s ss f)
  have refl: "snd (lpo pr n s s)" by (rule lpo_nstri_refl)
  from supt_arg refl have fact: "list_ex (\<lambda> t. snd (lpo pr n t s)) ss" by (auto)
  thus ?case 
    by (cases s, auto)
next
  case (supt_of_arg s ss t f)
  from supt_of_arg lpo_stri_imp_nstri[of "pr" n s t] have "\<exists> s \<in> set ss. snd (lpo pr n s t)" by auto
  hence "list_ex (\<lambda> s. snd (lpo pr n s t)) ss" by (auto)
  thus ?case
    by (cases t, auto)
qed

lemma lpo_nstri_mono: assumes rel: "snd (lpo pr n s t)"
       shows "snd (lpo pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  let ?len = "Suc (length bef + length aft)"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (lpo pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (lpo pr n ?s b)"
      proof (rule supt_imp_lpo_stri, cases "b \<in> set bef", (auto)[1], cases "b \<in> set aft", (auto)[1])
	assume "b \<notin> set bef" and "b \<notin> set aft"
	thus "?s \<rhd> b" using elem by blast
      qed
  qed
  from rel have "list_ex (\<lambda> s. snd (lpo pr n s t)) ?ss" by (auto)
  hence t: "fst (lpo pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). lpo pr n s t)"
  have lex_ext: "snd (lex_ext (lpo pr n) n ?ss ?ts)"
  proof (rule all_nstri_imp_lex_nstri[of ?ts _ ?ss, simplified], intro allI impI)
    fix i
    assume i: "i < ?len"
    show "snd (lpo pr n (?ss ! i) (?ts ! i))" (is "?nstri i")
    proof (cases "i = length bef")
      case True
      with rel show "?nstri i" by auto
    next
      case False
      show ?thesis 
      proof (cases "i < length bef")
        case True
        with lpo_nstri_refl[of "pr" n] show "?nstri i" by (simp add: nth_append)
      next
        case False
        with `i \<noteq> length bef` have "i - length bef > 0" by arith 
        then obtain j where i: "i - length bef = Suc j" by (cases "i - length bef", auto)
        with False lpo_nstri_refl[of "pr" n "aft ! j"] show "?nstri i" by (simp add: nth_append)
      qed
    qed
  qed
  show ?thesis by (simp only: lpo.simps, cases "list_ex (\<lambda> s. snd (lpo pr n s (Fun f ?ts))) ?ss", simp, simp add: t aft_bef lex_ext[simplified])
qed

lemma lpo_stri_mono: assumes rel: "fst (lpo pr n s t)"
       shows "fst (lpo pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (lpo pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (lpo pr n ?s b)"
      proof (rule supt_imp_lpo_stri, cases "b \<in> set bef", (auto)[1], cases "b \<in> set aft", (auto)[1])
	assume "b \<notin> set bef" and "b \<notin> set aft"
	thus "?s \<rhd> b" using elem by blast
      qed
  qed
  from lpo_stri_imp_nstri[OF rel] have "list_ex (\<lambda> s. snd (lpo pr n s t)) ?ss" by (auto)
  hence t: "fst (lpo pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). lpo pr n s t)"
  obtain stri nstri where st: "lpo pr n s t = (stri,nstri)" by force
  with rel have stri: "stri = True" by simp
  have lex_ext: "fst (lex_ext (lpo pr n) n ?ss ?ts)"
  proof (simp add: lex_ext_iff, rule exI[of _ "length bef"], simp add: st stri, intro allI impI)
    fix i
    assume i: "i < length bef"
    thus "snd (lpo pr n (?ss ! i) (?ts ! i))" by (simp add: lpo_nstri_refl[of "pr" n] nth_append)
  qed
  show ?thesis by (simp only: lpo.simps, cases "list_ex (\<lambda> s. snd (lpo pr n s (Fun f ?ts))) ?ss", simp, simp add: t aft_bef lex_ext)
qed


lemma lpo_stable: 
   "(fst (lpo pr n s t) \<longrightarrow> fst (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (lpo pr n s t) \<longrightarrow> snd (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"
proof (induct rule: lpo.induct[of "\<lambda> pr n s t. (fst (lpo pr n s t) \<longrightarrow> fst (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (lpo pr n s t) \<longrightarrow> snd (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"])
  case (1 "pr" n x y)
  show ?case by (auto simp: lpo_nstri_refl)
next
  case (2 "pr" n x g ts)
  show ?case by (cases "\<sigma>", cases "get_subst \<sigma> x", auto simp: Let_def all_nstri_imp_lex_nstri[of "[]", simplified])
next
  case (3 "pr" n f ss y)
  {
    assume "list_ex (\<lambda> s. snd (lpo pr n s (Var y))) ss"
    from this obtain s where s: "s \<in> set ss" and "snd (lpo pr n s (Var y))" by (force)
    with 3 have "snd (lpo pr n (s \<cdot> \<sigma>) (Var y \<cdot> \<sigma>))" by auto
    with s have "\<exists> s' \<in> set (map (\<lambda> s'. s' \<cdot> \<sigma>) ss). snd (lpo pr n s' (Var y \<cdot> \<sigma>))" by auto
    hence "list_ex (\<lambda> s'. snd (lpo pr n s' (Var y \<cdot> \<sigma>))) (map (\<lambda> s'. s' \<cdot> \<sigma>) ss)" by (force)
  }
  thus ?case by (cases "Var y \<cdot> \<sigma>", auto simp: Let_def)
next
  case (4 "pr" n f ss g ts)
  let ?map = "map (\<lambda> s'. s' \<cdot> \<sigma>)"
  let ?t = "Fun g ts"
  let ?s = "Fun f ss"
  let ?ssig = "Fun f ss \<cdot> \<sigma>"
  let ?tsig = "Fun g ts \<cdot> \<sigma>"
  let ?ss = "?map ss"
  let ?ts = "?map ts"
  show ?case
  proof (cases "\<exists> s \<in> set ss. snd (lpo pr n s ?t)")
    case True
    from this obtain s where s: "s \<in> set ss" and "snd (lpo pr n s ?t)" by auto
    with 4 have "snd (lpo pr n (s \<cdot> \<sigma>) ?tsig)" by auto
    with s have "\<exists> s' \<in> set ?ss. snd (lpo pr n s' ?tsig)" by auto
    hence "list_ex (\<lambda> s'. snd (lpo pr n s' ?tsig)) ?ss" by (force)
    thus ?thesis by auto
  next
    case False
    hence not1: "\<not> (list_ex (\<lambda> s. snd (lpo pr n s ?t)) ss)" by (force)
    show ?thesis 
    proof (cases "snd (lpo pr n ?s ?t)")
      case False
      with lpo_stri_imp_nstri have "\<not> (fst (lpo pr n ?s ?t))" by blast
      with False show ?thesis by auto
    next
      case True
      with not1 have smallTs: "\<forall> t \<in> set ts. fst (lpo pr n ?s t)" by (cases ?thesis, auto)
      with 4 not1 have "\<forall> t \<in> set ts. fst (lpo pr n ?ssig (t \<cdot> \<sigma>))" by auto
      hence smallTssig: "\<forall> t \<in> set ?ts. fst (lpo pr n ?ssig t)" by auto
      from True smallTs have ge: "pr f (length ss) \<ge> pr g (length ts)" by (cases ?thesis, auto simp: Let_def not1 simp del: list_all_ex)
      show ?thesis 
      proof (cases "pr f (length ss) > pr g (length ts)")
	case True
	with smallTssig not1 show ?thesis by auto
      next
	case False
	with ge have eq: "pr f (length ss) = pr g (length ts)" by auto
	with not1 smallTs True have snd_lex: "snd (lex_ext (lpo pr n) n ss ts)" by auto
	let ?ind = "\<lambda> i. (fst (lpo pr n (ss ! i) (ts ! i)) \<longrightarrow> fst (lpo pr n (ss ! i \<cdot> \<sigma>) (ts ! i \<cdot> \<sigma>))) \<and> (snd (lpo pr n (ss ! i) (ts ! i)) \<longrightarrow> snd (lpo pr n (ss ! i \<cdot> \<sigma>) (ts ! i \<cdot> \<sigma>)))"
        from 4(3)[OF not1,  of "pr f (length ss)" "pr g (length ts)", simplified, simplified eq, simplified, OF smallTs]
	have ind: "\<And> i. \<lbrakk>i < length ss; i < length ts\<rbrakk> \<Longrightarrow> ?ind i" by auto
        from snd_lex have eq_small: "length ss = length ts \<or> length ts \<le> n" by (simp add: lex_ext_iff)
        from ind snd_lex have snd_lex2: "snd (lex_ext (lpo pr n) n ?ss ?ts)"
          by (simp add: lex_ext_iff eq_small, auto)
	have nstri: "snd (lpo pr n ?ssig ?tsig)" by (simp add: Let_def eq snd_lex smallTssig[simplified] snd_lex2)
	{
	  assume "fst (lpo pr n ?s ?t)"
	  with not1 smallTs eq snd_lex have "fst (lex_ext (lpo pr n) n ss ts)" by auto
          with ind have fst_lex2: "fst (lex_ext (lpo pr n) n ?ss ?ts)"
            by (simp add: lex_ext_iff eq_small, auto)
	  have "fst (lpo pr n ?ssig ?tsig)" by (simp add: Let_def eq snd_lex smallTssig[simplified] fst_lex2)
	}
	with nstri show ?thesis by simp
      qed
    qed
  qed
qed

lemma lpo_least_1: assumes "pr f 0 = 0" 
  shows "snd (lpo pr n t (Fun f []))"
proof (cases t, simp add: assms)
  case (Fun g ts)
  with assms show ?thesis by (simp add: lex_ext_least_1)
qed

lemma lpo_least_2: assumes "pr f 0 = 0" 
  shows "\<not> fst (lpo pr n (Fun f []) t)"
proof (cases t, simp)
  case (Fun g ts)
  with assms show ?thesis by (simp add: Let_def lex_ext_least_2)
qed

lemma lpo_least_3: assumes "pr f 0 = 0" 
  and "snd (lpo pr n (Fun f []) t)"
  shows "snd (lpo pr n u t)"
using assms
proof (cases t, simp)
  case (Fun g ts)
  show ?thesis
  proof (cases ts)
    case Nil
    with Fun assms have "pr g 0 = 0" by (cases ?thesis, simp, simp add: Let_def)
    thus ?thesis by (simp only: Fun Nil, rule lpo_least_1)
  next
    case (Cons s tts)
    have "\<not> fst (lpo pr n (Fun f []) s)" by (rule lpo_least_2, rule `pr f 0 = 0`)
    with `snd (lpo pr n (Fun f []) t)` Fun Cons have False by (simp del: list_all_iff add: Let_def)
    thus ?thesis ..
  qed
qed

lemma lpo_compat[rule_format]: 
  fixes s t u :: "('f,'v)term"
  shows "(snd (lpo pr n s t) \<and> fst (lpo pr n t u) \<longrightarrow> fst (lpo pr n s u)) \<and> (fst (lpo pr n s t) \<and> snd (lpo pr n t u) \<longrightarrow> fst (lpo pr n s u)) \<and> (snd (lpo pr n s t) \<and> snd (lpo pr n t u) \<longrightarrow> snd (lpo pr n s u))" (is "?tran s t u")
proof (rule
    wf_induct[OF wf_measures[of "[\<lambda> (s,t,u). size s, \<lambda> (s,t,u). size t, \<lambda> (s,t,u). size u]"], 
    of "\<lambda> (s,t,u). ?tran s t u" "(s,t,u)", simplified], clarify)
  fix s t u :: "('f,'v)term"
  assume ind: "\<forall> s' t' u'. (size s' < size s \<longrightarrow> ?tran s' t' u') \<and> (size s' = size s \<and> (size t' < size t \<or> size t' = size t \<and> size u' < size u) \<longrightarrow> ?tran s' t' u')"
  let ?GR = "\<lambda> s t. fst (lpo pr n s t)"
  let ?GE = "\<lambda> s t. snd (lpo pr n s t)"
  show "?tran s t u"
  proof (cases t)
    case (Var x)
    { 
      assume "?GR t u"
      with Var have False by (cases u, auto)
    } 
    moreover
    {
      assume gr: "?GR s t" and ge: "?GE t u"
      have "?GR s u"
      proof (cases u)
	case (Var y)
	with ge `t = Var x` have "t = u" by auto
	with gr show ?thesis by auto
      next
	case (Fun h us)
	with ge Var have us: "us = []" and pri: "pr h 0 = 0" by auto
	from gr Var obtain f ss where s: "s = Fun f ss" by (cases s, auto)
	with gr Var obtain s' where s': "s' \<in> set ss" by (auto simp: Let_def)
	have "?GE s' u" by (simp only: Fun us, rule lpo_least_1, rule pri)
	with s' have "list_ex (\<lambda> s'. ?GE s' u) ss" by (auto)
	with s Fun show ?thesis by simp
      qed
    }
    moreover
    {
      assume ge1: "?GE s t" and ge2: "?GE t u"
      have "?GE s u"
      proof (cases u)
	case (Var y)
	with ge2 `t = Var x` have "t = u" by auto
	with ge1 show ?thesis by auto
      next
	case (Fun h us)
	with ge2 Var have us: "us = []" and pri: "pr h 0 = 0" by auto
	show ?thesis by (simp only: Fun us, rule lpo_least_1, rule pri) 
      qed
    }
    ultimately show ?thesis by blast
  next
    case (Fun g ts)
    let ?t = "Fun g ts"
    from Fun have t: "t = ?t" .
    show ?thesis 
    proof (cases s)
      case (Var x)
      {
	assume gr: "?GR s t"
	with Var Fun have False by auto
      }
      moreover
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	with Var Fun have pri: "pr g 0 = 0" and "ts = []" by auto
	with gr Fun have False using lpo_least_2[of "pr" g n u] by auto
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u"
	with Var Fun have pri: "pr g 0 = 0" and empty: "ts = []" by auto
	have "?GE s u" by (rule lpo_least_3[of "pr" g n u], rule pri, simp add: empty[symmetric] Fun[symmetric], rule ge2)
      }
      ultimately show ?thesis by blast
    next
      case (Fun f ss)
      let ?s = "Fun f ss"
      from Fun have s: "s = ?s" .
      let ?s1 = "\<exists> s' \<in> set ss. ?GE s' t"
      let ?t1 = "\<exists> t' \<in> set ts. ?GE t' u"
      let ?ls = "length ss"
      let ?lt = "length ts"
      let ?pf = "pr f ?ls"
      let ?pg = "pr g ?lt"
      term ?tran
      let ?tran2 = "\<lambda> a b c.  
         (snd (lpo pr n a b) \<and> fst (lpo pr n b c) \<longrightarrow> fst (lpo pr n a c)) \<and>
         (fst (lpo pr n a b) \<and> snd (lpo pr n b c) \<longrightarrow> fst (lpo pr n a c)) \<and> 
         (snd (lpo pr n a b) \<and> snd (lpo pr n b c) \<longrightarrow> snd (lpo pr n a c)) \<and> 
         (fst (lpo pr n a b) \<and> fst (lpo pr n b c) \<longrightarrow> fst (lpo pr n a c))"
      from s have "\<forall> s' \<in> set ss. size s' < size s" by auto
      with ind have ind2: "\<And> s' t' u'. \<lbrakk>s' \<in> set ss\<rbrakk> \<Longrightarrow> ?tran s' t' u'" by blast
      with lpo_stri_imp_nstri have ind3: "\<And> us s' t' u'. \<lbrakk>s' \<in> set ss; t' \<in> set ts; u' \<in> set us\<rbrakk> \<Longrightarrow> ?tran2 s' t' u'" by blast
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' ind2[of s' t u, simplified] ges gr have "?GR s' u" by auto
	  hence "?GE s' u" by (rule lpo_stri_imp_nstri)
	  with s' s show ?thesis by (cases u, auto simp: Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    from t' t False ge s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified] show "?GR s u" by force
	  next
	    case False
	    from t this gr obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	    let ?u = "Fun h us"	    
	    from s t u ge gr have ge: "?GE ?s ?t" and gr: "?GR ?t ?u" by auto
	    from `\<not> ?s1` t ge have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	    from `\<not> ?t1` u gr have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	    let ?lu = "length us"
	    let ?ph = "pr h ?lu"
	    from `\<not> ?s1` t ge st' have fg: "?pf \<ge> ?pg" by (cases ?thesis, auto simp: Let_def) 
	    from `\<not> ?t1` u gr tu' have fh: "?pg \<ge> ?ph" by (cases ?thesis, auto simp: Let_def) 
	    from tu' s t ge spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" by force
	    show ?thesis
	    proof (cases "?pf > ?pg \<or> ?pg > ?ph")
	      case True
	      with fg fh have "?pf > ?ph" by auto
	      with su' s u show ?thesis by (auto)
	    next
	      case False              
	      from False fg fh have fg: "?pf = ?pg" and gh: "?pg = ?ph" and fh: "?pf = ?ph" by auto
	      from fg `\<not> ?s1` t ge st' have lex1: "snd (lex_ext (lpo pr n) n ss ts)" by auto
	      from gh `\<not> ?t1` u gr tu' have lex2: "fst (lex_ext (lpo pr n) n ts us)" by auto
	      from lex1 lex2 lex_ext_compat[OF ind3, of ss ts us] have "fst (lex_ext (lpo pr n) n ss us)" by auto
	      with s u fh su' show ?thesis by auto
	    qed
	  qed
	qed
      }
      moreover
      {
	assume gr: "?GR s t" and ge: "?GE t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' ind2[of s' t u, simplified] ges ge have "?GE s' u" by auto
	  with s' s show ?thesis by (cases u, auto simp: Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    have tt': "?GE t t'" by (rule lpo_stri_imp_nstri[OF supt_imp_lpo_stri], auto simp: t t')
	    from t' t False gr s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified] show "?GR s u" by force
	  next
	    case False
	    from t this ge obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	    let ?u = "Fun h us"	    
	    from s t u ge gr have gr: "?GR ?s ?t" and ge: "?GE ?t ?u" by auto
	    from `\<not> ?s1` t gr have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	    from `\<not> ?t1` u ge have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	    let ?lu = "length us"
	    let ?ph = "pr h ?lu"
	    from `\<not> ?s1` t gr st' have fg: "?pf \<ge> ?pg" by (cases ?thesis, auto simp: Let_def) 
	    from `\<not> ?t1` u ge tu' have fh: "?pg \<ge> ?ph" by (cases ?thesis, auto simp: Let_def) 
	    from tu' s t lpo_stri_imp_nstri[OF gr] spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" by force
	    show ?thesis
	    proof (cases "?pf > ?pg \<or> ?pg > ?ph")
	      case True
	      with fg fh have "?pf > ?ph" by auto
	      with su' s u show ?thesis by (auto)
	    next
	      case False
	      with fg fh have fg: "?pf = ?pg" and gh: "?pg = ?ph" and fh: "?pf = ?ph" by auto
	      from fg `\<not> ?s1` t gr st' have lex1: "fst (lex_ext (lpo pr n) n ss ts)" by auto
	      from gh `\<not> ?t1` u ge tu' have lex2: "snd (lex_ext (lpo pr n) n ts us)" by auto
	      from lex1 lex2 lex_ext_compat[OF ind3, of ss ts us]
	      have "fst (lex_ext (lpo pr n) n ss us)" by auto
	      with s u fh su' show ?thesis by (auto)
	    qed
	  qed
	qed
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u" and ngt1: "\<not> ?GR s t" and ngt2: "\<not> ?GR t u"
	have "\<not> ?s1"
	proof
	  assume ?s1
	  with s have "?GR s t" by (cases t, auto)
	  with ngt1 show False ..
	qed
	have "\<not> ?t1"
	proof
	  assume ?t1
	  with t have "?GR t u" by (cases u, auto)
	  with ngt2 show False ..
	qed	
	from t `\<not> ?t1` ge2 obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	let ?u = "Fun h us"	    
	from s t u ge1 ge2 have ge1: "?GE ?s ?t" and ge2: "?GE ?t ?u" by auto
	from `\<not> ?s1` t ge1 have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	from `\<not> ?t1` u ge2 have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	let ?lu = "length us"
	let ?ph = "pr h ?lu"
	from tu' s t ge1 spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" by force
	from `\<not> ?s1` st' ge1 ngt1 s t have fg: "?pf = ?pg" by (simp add: Let_def, cases ?thesis, simp, cases "?pf > ?pg", auto) 
	from `\<not> ?t1` tu' ge2 ngt2 t u have gh: "?pg = ?ph" by (simp add: Let_def, cases ?thesis, simp, cases "?pg > ?ph", auto)
	from `\<not> ?s1` st' ge1 s t fg have lex1: "snd (lex_ext (lpo pr n) n ss ts)" by auto
	from `\<not> ?t1` tu' ge2 t u gh have lex2: "snd (lex_ext (lpo pr n) n ts us)" by auto
	from lex1 lex2 lex_ext_compat[OF ind3, of ss ts us]
	have "snd (lex_ext (lpo pr n) n ss us)" by auto
	with fg gh su' s u have "?GE s u" by auto
      } 
      ultimately
      show ?thesis using lpo_stri_imp_nstri[of "pr" n s u] by auto
    qed
  qed
qed


lemma lpo_trans: assumes one: "fst (lpo pr n s t)" and two: "fst (lpo pr n t u)" shows "fst (lpo pr n s u)"
  using assms lpo_compat lpo_stri_imp_nstri[OF one] by blast

definition LPO_S where "LPO_S pr n = {(s,t) . fst (lpo pr n s t)}"
definition LPO_NS where "LPO_NS pr n = {(s,t). snd (lpo pr n s t)}"

lemma LPO_NS_refl: "refl (LPO_NS pr n)"
unfolding refl_on_def LPO_NS_def
 by (simp add: lpo_nstri_refl)

lemma LPO_NS_trans: "trans (LPO_NS pr n)"
unfolding trans_def LPO_NS_def 
using lpo_compat by blast

lemma LPO_S_trans: "trans (LPO_S pr n)"
unfolding trans_def LPO_S_def 
using lpo_trans by blast

lemma LPO_compat : "LPO_NS pr n O LPO_S pr n \<subseteq> LPO_S pr n"
  using lpo_compat[of "pr" n] 
  by (unfold LPO_S_def LPO_NS_def, auto)

abbreviation lpo_prop where "lpo_prop pr n X \<equiv> \<forall> t. (\<forall> s. fst (lpo pr n t s) \<longrightarrow> s \<in> X) \<longrightarrow> t \<in> X"

(* note that the properties of W W1 W2 do not depend on the definition of lpo, they can
   be generalized. However, due to lazyness, this is currently not done *)
definition W :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f,'v)term set" where "W pr n \<equiv> \<Inter> { X. lpo_prop pr n X}"

lemma W1 : "(t \<in> W pr n) = ( \<forall> s. fst (lpo pr n (t :: ('f,'v)term) s) \<longrightarrow> s \<in> W pr n)" (is "?r = ?l")
proof
  assume ?l
  show "t \<in> W pr n"
    unfolding W_def
  proof (clarify)
    fix X :: "('f,'v)term set"
    assume im: "lpo_prop pr n X"
    hence subset: "W pr n \<subseteq> X" unfolding W_def by blast
    from im have imp: "(\<forall> s. fst (lpo pr n t s) \<longrightarrow> s \<in> X) \<longrightarrow> t \<in> X" ..
    have "\<forall> s. fst (lpo pr n t s) \<longrightarrow> s \<in> X" using `?l` subset by blast
    with imp show "t \<in> X" by blast
  qed
next
  assume ?r
  show ?l 
  proof (intro allI impI)
    fix s :: "('f,'v)term"
    assume lpo: "fst (lpo pr n t s)"
    show "s \<in> W pr n" unfolding W_def
    proof (clarify)
      fix X :: "('f,'v)term set"
      assume imp: "lpo_prop pr n X"
      show "s \<in> X" 
      proof (rule ccontr)
	assume "s \<notin> X"
	hence "lpo_prop pr n (X - {t})" using imp lpo by blast
	with `?r` show False unfolding W_def by blast
      qed
    qed
  qed
qed

	
fun W2_fun :: "(('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> ('f,'v)term"
  where "W2_fun P pr n 0 = (SOME x . x \<in> W pr n \<and> \<not> P x)"
  | "W2_fun P pr n (Suc m) = (SOME x. x \<in> W pr n \<and> \<not> P x \<and> fst (lpo pr n (W2_fun P pr n m) x))"

lemma W2[rule_format] : "\<forall> t \<in> W pr m. ( \<forall> s. fst (lpo pr m (t :: ('f,'v)term) s) \<longrightarrow> P s) \<longrightarrow> P t \<Longrightarrow> \<forall> t \<in> W pr m. P t"
proof (rule ccontr)
  assume one: "\<not> (\<forall> t \<in> W pr m. P t)"
    and other: "\<forall> (t :: ('f,'v)term) \<in> W pr m. (\<forall> s. fst (lpo pr m t s) \<longrightarrow> P s) \<longrightarrow> P t"
  have "\<forall> n. W2_fun P pr m n \<in> W pr m \<and> \<not> P (W2_fun P pr m n) \<and> (n > 0 \<longrightarrow> fst (lpo pr m (W2_fun P pr m (n - 1)) (W2_fun P pr m n)))" (is "\<forall> n. ?P n")
  proof
    fix n
    show "?P n"
    proof (induct n)
      case 0
      from one obtain t where tW: "t \<in> W pr m \<and> \<not> P t" by auto 
      show ?case by (simp, rule someI[of "\<lambda> t. t \<in> W pr m \<and> \<not> P t" t], rule tW)
    next
      case (Suc n)
      with other obtain s where lpo: "fst (lpo pr m (W2_fun P pr m n) s) \<and> \<not> P s" by blast
      from Suc W1[of "W2_fun P pr m n" "pr" m] lpo have "s \<in> W pr m" by blast
      with lpo have conj: "s \<in> W pr m \<and> \<not> P s \<and> fst (lpo pr m (W2_fun P pr m n) s)" by simp
      from someI[of "\<lambda> s. s \<in> W pr m \<and> \<not> P s \<and> fst (lpo pr m (W2_fun P pr m n) s)" s, OF conj]
      show ?case by simp
    qed
  qed
  hence all: "\<forall> n. W2_fun P pr m n \<in> W pr m \<and> fst (lpo pr m (W2_fun P pr m n) (W2_fun P pr m (Suc n)))"
    by auto
  from this obtain g :: "nat \<Rightarrow> ('f,'v)term" where g: "(\<forall> n. g n \<in> W pr m \<and> fst (lpo pr m (g n) (g (Suc n))))" by blast 
  from g have "lpo_prop pr m (W pr m - range g)" unfolding W_def by blast
  with g show False unfolding W_def by blast
qed

hide_const W2_fun


lemma W_imp_W_next: assumes "t \<in> W pr m" and "fst (lpo pr m t s)"
  shows "s \<in> W pr m"
  using assms W1[of t] by blast


lemma W_imp_SN : fixes "pr" :: "'f \<Rightarrow> nat \<Rightarrow> nat" and t :: "('f,'v)term"
  assumes t: "t \<in> W pr m" shows "SN_elt (LPO_S pr m) t"
proof (rule ccontr)
  assume "\<not> ?thesis"
  from this obtain g where g0: "g 0 = t" and steps: "\<forall> n. fst (lpo pr m (g n) (g (Suc n)))" unfolding SN_elt_def[of "LPO_S pr m" t] LPO_S_def by force
  have g1: "\<forall> n. g n \<in> W pr m" 
  proof
    fix n
    show "g n \<in> W pr m"
    proof (induct n, simp add: g0 t)
      case (Suc n)
      thus ?case by (rule W_imp_W_next[OF _ spec[OF steps, of n]])
    qed
  qed
  with steps have "lpo_prop pr m (W pr m - range g)" unfolding W_def by blast
  with g1 show False unfolding W_def by blast
qed


lemma var_in_W : "Var x \<in> W pr n"
proof (simp only: W1[of "Var x" "pr" n], intro allI impI)
  fix s
  assume "fst (lpo pr n (Var x) s)"
  hence False by (cases s, auto)
  thus "s \<in> W pr n" ..
qed

lemma nstri_and_W : assumes t: "t \<in> W pr n" and nstri: "snd (lpo pr n t s)"
  shows "s \<in> W pr n"
proof (simp add: W1[of s], intro impI allI)
  fix u
  assume "fst (lpo pr n s u)"
  with lpo_compat nstri have u: "fst (lpo pr n t u)" by blast
  with t W1[of t] show "u \<in> W pr n" by auto
qed

definition W_list : "W_list pr m ts \<equiv> \<forall> t \<in> set ts. t \<in> W pr m"


  
lemma fun_in_W[rule_format] : fixes "pr" :: "'f \<Rightarrow> nat \<Rightarrow> nat" and ts :: "('f,'v)term list"
  assumes "W_list pr m ts"
  shows "Fun f ts \<in> W pr m"
proof -
  let ?cond = "\<lambda> ys xs :: ('f,'v)term list.(\<forall> y \<in> set ys. SN_elt (LPO_S pr m) y) \<and> fst (lex_ext (lpo pr m) m ys xs)"
  let ?r4 = "{(ys :: ('f,'v)term list,xs). ?cond ys xs}"
  let ?r3 = "{(xs :: ('f,'v)term list,ys). ?cond ys xs}"
  let ?r2 = "{(xs :: ('f,'v)term list,ys). W_list pr m ys \<and> fst (lex_ext (lpo pr m) m ys xs)}"
  let ?r1 = "{(x :: nat,y). x < y} <*lex*> ?r2"
  let ?m = "\<lambda> (f,ts). (pr f (length ts), ts)"
  let ?r = "{(a,b). (?m a, ?m b) \<in> ?r1}"
  let ?gr = "\<lambda> t s. fst (lpo pr m t s)"
  let ?ge = "\<lambda> t s. snd (lpo pr m t s)"
  have supset: "?r3 \<supseteq> ?r2" unfolding W_list using W_imp_SN[of _ "pr" m] by auto
  have "wf ?r2" 
  proof (rule wf_subset[OF _ supset])
    have id: "?r4^-1 = ?r3" by auto
    have "(SN :: ('f,'v)term list ars \<Rightarrow> bool) ?r4" 
      unfolding LPO_S_def      
    proof (rule lex_ext_SN[of "lpo pr m" m], simp add: lpo_stri_imp_nstri)
      fix x y z 
      assume "?ge x y" and "?gr y z"
      thus "?gr x z" using lpo_compat[of "pr" m x y z] by simp
    qed
    thus "wf ?r3" by (simp only: SN_iff_wf[of ?r4] id)
  qed
  hence wf_r1: "wf ?r1" by (rule wf_lex_prod[OF wf_less, of ?r2])
  have wf_r: "wf ?r" using wf_inv_image[OF wf_r1, of ?m] unfolding inv_image_def .
  let ?W = "W pr m"
  let ?WL = "W_list pr m"
  let ?ind = "\<lambda> (f,ts).  ?WL ts \<longrightarrow> Fun f ts \<in> ?W"
  have "?ind (f,ts)" 
  proof (rule wf_induct[OF wf_r, of ?ind])
    fix fts
    assume ind: "\<forall> gss. (gss,fts) \<in> ?r \<longrightarrow> ?ind gss"
    show "?ind fts"
    proof (cases fts)
      case (Pair f ts)
      thus ?thesis
      proof (simp, intro impI)
	assume ts: "?WL ts"
	let ?t = "Fun f ts"	
	show "?t \<in> ?W"
	proof (simp only: W1[of ?t], intro allI impI)
	  fix s
	  assume "?gr ?t s"
	  thus "s \<in> ?W"
	  proof (induct s, simp add: var_in_W)
	    case (Fun g ss)
	    let ?s = "Fun g ss"
	    from Fun have t_gr_s: "?gr ?t ?s" by auto
	    show "?s \<in> ?W"
	    proof (cases "\<exists> t \<in> set ts. ?ge t ?s")
	      case True
	      then obtain t where "t \<in> set ts" and ge: "?ge t ?s" by auto	      
	      with ts have "t \<in> ?W" unfolding W_list by auto
	      show "?s \<in> ?W" by (rule nstri_and_W[OF `t \<in> ?W` ge])
	    next
	      case False
	      let ?ls = "length ss"
	      let ?lt = "length ts"
	      let ?pf = "pr f ?lt"
	      let ?pg = "pr g ?ls"
	      from False t_gr_s have f_ge_g: "?pf \<ge> ?pg" by (cases ?thesis, auto simp: Let_def)
	      from False t_gr_s f_ge_g have small_ss: "\<forall> s \<in> set ss. ?gr ?t s" by (cases ?thesis, auto simp: Let_def)
	      with Fun have ss_W: "?WL ss" unfolding W_list by auto
	      show ?thesis
	      proof (cases "?pf = ?pg")
		case False
		with f_ge_g have f_gt_g: "?pf > ?pg" by auto
		hence "((g,ss),(f,ts)) \<in> ?r" by (simp add: in_lex_prod)
		with ind have "?ind (g,ss)" using Pair by auto
		with ss_W show ?thesis by auto
	      next
		case True
		from True False t_gr_s small_ss have lex: "fst (lex_ext (lpo pr m) m ts ss)" by (cases ?thesis, auto simp: Let_def)
		from True lex ts have "((g,ss),(f,ts)) \<in> ?r" by (simp add: in_lex_prod)
		with ind have "?ind (g,ss)" using Pair by auto
		with ss_W show ?thesis by auto
	      qed
	    qed
	  qed
	qed
      qed
    qed
  qed
  with assms show ?thesis by auto
qed


lemma W_is_univ: "t \<in> W pr m"
  by (induct t, auto simp: var_in_W fun_in_W W_list)


lemma LPO_S_SN: "SN (LPO_S pr n)" 
  unfolding SN_def by (intro allI, rule W_imp_SN, rule W_is_univ)

hide_const W W_list

interpretation LPO_redpair : redpair_order "LPO_S pr n" "LPO_NS pr n"
proof (unfold_locales, rule LPO_S_SN, rule LPO_compat)
  show "subst_closed (LPO_NS pr n)"
    unfolding subst_closed_def 
  proof (rule subsetI2)
    fix s t
    assume "(s,t) \<in> CS (LPO_NS pr n)"
    then obtain u v \<sigma> where "(u,v) \<in> LPO_NS pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: subst_closure.cases) auto
    thus "(s,t) \<in> LPO_NS pr n" unfolding LPO_NS_def using lpo_stable[of "pr" n] by auto
  qed
next
  show "subst_closed (LPO_S pr n)"
    unfolding subst_closed_def 
  proof (rule subsetI2)
    fix s t
    assume "(s,t) \<in> CS (LPO_S pr n)"
    then obtain u v \<sigma> where "(u,v) \<in> LPO_S pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: subst_closure.cases) auto
    thus "(s,t) \<in> LPO_S pr n" unfolding LPO_S_def using lpo_stable[of "pr" n] by auto
  qed
next
  show "one_step_ctxt_closed (LPO_NS pr n)"
    unfolding one_step_ctxt_closed_def LPO_NS_def
    by (intro allI impI, simp del: lpo.simps, rule lpo_nstri_mono)
next
  show "refl (LPO_NS pr n)" by (rule LPO_NS_refl)
next
  show "trans (LPO_NS pr n)" by (rule LPO_NS_trans)
next
  show "trans (LPO_S pr n)" by (rule LPO_S_trans)
qed 

definition
  lpo_strict
where
  "lpo_strict pr n \<equiv> \<lambda>(s, t). check (fst (lpo pr n s t)) (''could not orient '' +#+ shows s +@+ '' >LPO '' +#+ shows t +@+ shows_nl)"

definition
  lpo_nstrict
where
  "lpo_nstrict pr n \<equiv> \<lambda> (s,t). check (snd (lpo pr n s t)) (''could not orient '' +#+ shows s +@+ '' >=LPO '' +#+ shows t +@+ shows_nl)"

interpretation LPO_redpair_impl : redpair_impl "LPO_S pr n" "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n"
  by (unfold_locales, unfold LPO_S_def LPO_NS_def lpo_strict_def lpo_nstrict_def, simp_all add: split_paired_all) 


interpretation ce_redpair "LPO_S pr n" "LPO_NS pr n"
proof(unfold_locales, intro exI allI impI)
  fix m c d
  show "ce_trs (c,d,m) \<subseteq> LPO_NS pr n"
    unfolding LPO_NS_def 
    by (simp add: ce_trs.simps, auto simp: lpo_stri_imp_nstri[OF supt_imp_lpo_stri])
qed

interpretation LPO_ce_redpair : ce_redpair_impl "LPO_S pr n" "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n" ..

interpretation mono_ce_redpair "LPO_S pr n" "LPO_NS pr n" 
proof(unfold_locales, intro exI allI impI)
  fix m c d
  show "ce_trs (c,d,m) \<subseteq> LPO_S pr n"
    unfolding LPO_S_def 
    by (simp add: ce_trs.simps, auto simp: supt_imp_lpo_stri)
qed

interpretation LPO_ce_mono_redpair : ce_mono_redpair_impl "LPO_S pr n" "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n" ch
proof(unfold_locales)
  show "one_step_ctxt_closed (LPO_S pr n)"
    unfolding one_step_ctxt_closed_def LPO_S_def
    by (intro allI impI, simp del: lpo.simps, rule lpo_stri_mono)
qed

(* more efficient implementation which does not require the bound *)

function lpo_unbounded :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool \<times> bool" where
  "lpo_unbounded _ (Var x) (Var y) = (False, x = y)"
| "lpo_unbounded pr (Var x) (Fun g ts) = (False, ts = [] \<and> pr g 0 = 0)"
| "lpo_unbounded pr (Fun f ss) (Var y) = (let con = list_ex (\<lambda> s. snd (lpo_unbounded pr s (Var y))) ss in (con,con))" 
| "lpo_unbounded pr (Fun f ss) (Fun g ts) = (
    if list_ex (\<lambda> s. snd (lpo_unbounded pr s (Fun g ts))) ss 
       then (True,True)
       else (let prl = pr f (length ss); prr = pr g (length ts) in 
         if prl \<ge> prr \<and> (list_all (\<lambda> t. fst (lpo_unbounded pr (Fun f ss) t)) ts)
          then if prl > prr then (True,True) 
            else lex_ext_unbounded (lpo_unbounded pr) ss ts
          else (False,False)))"
by pat_completeness auto

termination lpo_unbounded
proof (relation "measures [\<lambda> (pr,s,t). size s + size t]", (auto)[4])
  fix "pr" and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa xb xc i
  assume "i < length ss" and "i < length ts"
  then obtain s t where si: "s = ss ! i" and s: "s \<in> set ss" and ti: "t = ts ! i" and t: "t \<in> set ts" by auto
  from s have s1: "size s \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size t \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 si ti show "((pr, ss ! i, ts ! i), pr, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,s,t). size s + size t]"
    by auto
qed

lemma lpo_to_lpo_unbounded : assumes "\<forall> f i. (f, i) \<in> funas_term s \<union> funas_term t \<longrightarrow> i \<le> n" (is "?b s t")
  shows "lpo pr n s t = lpo_unbounded pr s t" (is "?e pr s t")
proof -
  let ?p = "\<lambda> u s t. ?b s t \<longrightarrow> lpo u n s t = lpo_unbounded u s t"
  {
    fix p s t
    have "?p p s t"
    proof (induct rule: lpo_unbounded.induct[where P = ?p], simp, simp)
      case (3 p f ss y)
      show ?case 
      proof (intro impI)
        assume "?b (Fun f ss) (Var y)"
        hence "\<And> s. s \<in> set ss \<Longrightarrow> ?b s (Var y)" by auto
        with 3 have "\<And> s. s \<in> set ss \<Longrightarrow> ?e p s (Var y)" by auto
        thus "?e p (Fun f ss) (Var y)"
          by (simp add: Let_def)
      qed
    next
      case (4 p f ss g ts)
      show ?case
      proof (intro impI)
        assume "?b (Fun f ss) (Fun g ts)"
        hence bs: "\<And> s. s \<in> set ss \<Longrightarrow> ?b s (Fun g ts)" and bt: "\<And> t. t \<in> set ts \<Longrightarrow> ?b (Fun f ss) t" and ss: "length ss \<le> n" and ts: "length ts \<le> n" by auto
        with 4 have s: "\<And> s. s \<in> set ss \<Longrightarrow> ?e p s (Fun g ts)" by auto
        show "?e p (Fun f ss) (Fun g ts)"
        proof (cases "list_ex (\<lambda> s. snd (lpo p n s (Fun g ts))) ss")
          case True
          with s have "list_ex (\<lambda> s. snd (lpo_unbounded p s (Fun g ts))) ss" by auto
          with True show ?thesis by auto
        next
          case False note oFalse = this
          with s have False2: "\<not> list_ex (\<lambda> s. snd (lpo_unbounded p s (Fun g ts))) ss" by auto
          with bt False 4(2) have t: "\<And> t. t \<in> set ts \<Longrightarrow> ?e p (Fun f ss) t" by force
          show ?thesis 
          proof (cases "p f (length ss) \<ge> p g (length ts) \<and> list_all (\<lambda> t. fst (lpo p n (Fun f ss) t)) ts")
            case False
            with t have t: "\<not> (p f (length ss) \<ge> p g (length ts) \<and> list_all (\<lambda> t. fst (lpo_unbounded p (Fun f ss) t)) ts)" by auto
            thus ?thesis by (simp add: oFalse False False2 Let_def t del: list_all_ex)
          next
            case True
            with t have t: "p f (length ss) \<ge> p g (length ts) \<and> list_all (\<lambda> t. fst (lpo_unbounded p (Fun f ss) t)) ts" by auto
            thus ?thesis proof (simp add: True False False2 Let_def t del: list_all_ex, intro impI,
              simp only: lex_ext_to_lex_ext_unbounded[OF ss ts, of "lpo p n"])
              assume p: "\<not> p g (length ts) < p f (length ss)"
              {
                fix i
                assume iss: "i < length ss" and its: "i < length ts"
                with 4(3)[OF False2 _ _ t p, simplified, OF iss its]
                bs[of "ss ! i"]  bt[of "ts ! i"] have "lpo p n (ss ! i) (ts ! i) = lpo_unbounded p (ss ! i) (ts ! i)" by auto
              } 
              thus "lex_ext_unbounded (lpo p n) ss ts = lex_ext_unbounded (lpo_unbounded p) ss ts"
                by (simp only: lex_ext_unbounded_iff, auto )
            qed
          qed
        qed
      qed
    qed
  }
  with assms 
  show ?thesis by simp
qed

definition
  lpo_strict_unbounded
where
  "lpo_strict_unbounded pr \<equiv> \<lambda>(s, t). check (fst (lpo_unbounded pr s t)) (''could not orient '' +#+ shows s +@+ '' >LPO '' +#+ shows t +@+ shows_nl)"

definition
  lpo_nstrict_unbounded
where
  "lpo_nstrict_unbounded pr \<equiv> \<lambda>(s, t). check (snd (lpo_unbounded pr s t)) (''could not orient '' +#+ shows s +@+ '' >=LPO '' +#+ shows t +@+ shows_nl)"

types 'f lpo_repr = "(('f \<times> nat) \<times> nat)list"


fun shows_lpo_repr :: "('f::show)lpo_repr \<Rightarrow> shows"
where "shows_lpo_repr prs = (''LPO with the following precedence'' +#+ shows_nl +@+ shows_map (\<lambda>((f,n),pr).
  ''precedence('' +#+ shows f +@+ ''['' +#+ shows n +@+ '']) = '' +#+ shows pr +@+ shows_nl
) prs +@+ shows_nl
+@+ ''precedence(_) = 0'' +#+ shows_nl
)"

definition full_af :: "'f af" where "full_af f n i \<equiv> True"

definition create_LPO_redpair :: "(('f :: show) lpo_repr \<Rightarrow> ('g \<Rightarrow> nat \<Rightarrow> nat)) \<Rightarrow> 'f lpo_repr \<Rightarrow> ('g :: show,'v :: show)redpair_impl"
where "create_LPO_redpair lpo_repr_to_pr pr \<equiv> let p = lpo_repr_to_pr pr in 
  (OK, lpo_strict_unbounded p, lpo_nstrict_unbounded p, full_af, OK, shows_lpo_repr pr)"

interpretation LPO_redpair: generic_redpair_impl "create_LPO_redpair (lpo_repr_to_pr :: ('f :: show)lpo_repr \<Rightarrow> ('g :: show \<Rightarrow> nat \<Rightarrow> nat))"
proof
  fix "pr" :: "('f :: show)lpo_repr" and s_list ns_list :: "('g :: show,'v :: show)rule list"
  let ?rp = "create_LPO_redpair lpo_repr_to_pr pr :: ('g,'v)redpair_impl"
  let ?pr = "lpo_repr_to_pr pr"
  let ?af = "rp_af ?rp :: ('g af)"
  assume stri: "isOK(check_allm (rp_strict ?rp) s_list)" and nstri: "isOK (check_allm (rp_non_strict ?rp) ns_list)"
  obtain n where n: "n = max_list (map (\<lambda> (s,t). max_list (map snd ((funas_term_list s) @ (funas_term_list t)))) (s_list @ ns_list))" by auto
  {
    fix s t
    assume st: "(s,t) \<in> set (s_list @ ns_list)"
    have "\<forall> f i. (f,i) \<in> funas_term s \<union> funas_term t \<longrightarrow> i \<le> n"
    proof (intro allI impI)
      fix f i
      assume fi: "(f,i) \<in> funas_term s \<union> funas_term t"
      from max_list st have one: "max_list (map snd ((funas_term_list s) @ (funas_term_list t))) \<le> n" (is "?ml \<le> n") unfolding n by force
      from fi max_list have two: "i \<le> ?ml" by force
      from one two show "i \<le> n" by simp
    qed
    hence "lpo ?pr n s t = lpo_unbounded ?pr s t" by (rule lpo_to_lpo_unbounded)
  } note main = this
  have mono: "mono_ce_af_redpair (LPO_S ?pr n) (LPO_NS ?pr n) full_af"
    by (unfold_locales, unfold af_compatible_def, simp add: Let_def full_af_def)
  show "\<exists> S NS. ce_af_redpair S NS ?af  \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> (isOK (rp_mono ?rp) \<longrightarrow> mono_ce_af_redpair S NS ?af \<and> one_step_ctxt_closed S)"
  proof (simp add: create_LPO_redpair_def Let_def, intro exI, intro conjI, 
    rule mono_ce_af_redpair.axioms(1)[OF mono], simp_all add: mono)
    show "one_step_ctxt_closed (LPO_S ?pr n :: ('g,'v)term ars)"
      by (rule LPO_ce_mono_redpair.checkMonoSound[of OK], simp)
  next
    {
      fix s t
      assume st: "(s,t) \<in> set s_list"
      with stri have "fst (lpo_unbounded ?pr s t)" by (auto simp: create_LPO_redpair_def lpo_strict_unbounded_def Let_def)
      with st have "fst (lpo ?pr n s t)" by (simp add: main)
      hence "(s,t) \<in> LPO_S ?pr n" unfolding LPO_S_def by simp
    }
    thus "set s_list \<subseteq> LPO_S ?pr n" ..
  next
    {
      fix s t
      assume st: "(s,t) \<in> set ns_list"
      with nstri have "snd (lpo_unbounded ?pr s t)" by (auto simp: create_LPO_redpair_def lpo_nstrict_unbounded_def Let_def)
      with st have "snd (lpo ?pr n s t)" by (simp add: main)
      hence "(s,t) \<in> LPO_NS ?pr n" unfolding LPO_NS_def by simp
    }
    thus "set ns_list \<subseteq> LPO_NS ?pr n" ..
  qed
qed

(* and undo the list_all_ex[simp] *)
declare list_all_ex[simp del]

end
