(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPairImpl imports PolyCarrier RPO ArgumentFilter
begin


datatype 'f redpair = 
    Int_carrier "('f,int)lpoly_interL" 
  | Rat_carrier rat "('f,rat)lpoly_interL"
  | Arctic_carrier "('f,arctic)lpoly_interL" 
  | Arctic_rat_carrier rat "('f,arctic_rat)lpoly_interL" 
  | Int_mat_carrier nat nat "('f,int mat)lpoly_interL"
  | Rat_mat_carrier rat nat nat "('f,rat mat)lpoly_interL"
  | Arc_mat_carrier nat "('f,arctic mat)lpoly_interL"
  | Arc_rat_mat_carrier rat nat "('f,arctic_rat mat)lpoly_interL"
  | LPO "'f lpo_repr" "'f afs_list"

fun lpo_repr_to_pr :: "'f lpo_repr \<Rightarrow> ('f filtered \<Rightarrow> nat \<Rightarrow> nat)"
where "lpo_repr_to_pr prs (FPair f a) _ = (case lookup (f,a) prs of None \<Rightarrow> 0 | Some x \<Rightarrow> x)"


fun get_redpair :: "('f :: show) redpair \<Rightarrow> ('f,'v :: show)redpair_impl"
where 
  "get_redpair (Int_carrier I) = int_redpair I"
| "get_redpair (Rat_carrier d I) = rat_redpair d I"
| "get_redpair (Arctic_carrier I) = arctic_redpair I"
| "get_redpair (Arctic_rat_carrier d I) = arctic_rat_redpair d I"
| "get_redpair (Int_mat_carrier n sd I) = int_mat_redpair n sd I"
| "get_redpair (Rat_mat_carrier d n sd I) = rat_mat_redpair n sd d I"
| "get_redpair (Arc_mat_carrier n I) = arctic_mat_redpair n I"
| "get_redpair (Arc_rat_mat_carrier d n I) = arctic_rat_mat_redpair n d I"
| "get_redpair (LPO prec pi) = af_redpair pi (create_LPO_redpair lpo_repr_to_pr prec)"

lemma get_redpair : "generic_redpair_impl get_redpair"
proof
  fix rp :: "('f :: show)redpair" and s_list ns_list :: "('f,'v::show)rules"
  let ?rp = "get_redpair rp :: ('f,'v)redpair_impl"
  assume v: "isOK( rp_valid ?rp)" and S: "isOK(check_allm (rp_strict ?rp) s_list)" and NS: "isOK(check_allm (rp_non_strict ?rp) ns_list)"
  show "\<exists> S NS. ce_af_redpair S NS (rp_af ?rp) \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> (isOK(rp_mono ?rp) \<longrightarrow> mono_ce_af_redpair S NS (rp_af ?rp) \<and> one_step_ctxt_closed S)"
  proof (cases rp)
    case (Int_carrier I)
    hence id: "?rp = (int_redpair :: ('f,int) lpoly_interL \<Rightarrow> ('f,'v)redpair_impl) I" (is "_ = ?I I") by simp    
    have rp: "generic_redpair_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redpair_impl.generate_redpair[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id]]) 
  next
    case (Rat_carrier d I)
    hence id: "?rp = (rat_redpair :: rat \<Rightarrow> ('f,rat) lpoly_interL \<Rightarrow> ('f,'v)redpair_impl) d I" (is "_ = ?I I") by simp    
    have rp: "generic_redpair_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redpair_impl.generate_redpair[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id]]) 
  next
    case (Arctic_carrier I)
    hence id: "?rp = (arctic_redpair :: ('f,arctic) lpoly_interL \<Rightarrow> ('f,'v)redpair_impl) I" (is "_ = ?I I") by simp    
    have rp: "generic_redpair_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redpair_impl.generate_redpair[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id]]) 
  next
    case (Arctic_rat_carrier d I)
    hence id: "?rp = (arctic_rat_redpair :: rat \<Rightarrow> ('f,arctic_rat) lpoly_interL \<Rightarrow> ('f,'v)redpair_impl) d I" (is "_ = ?I I") by simp    
    have rp: "generic_redpair_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redpair_impl.generate_redpair[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id]]) 
  next
    case (Int_mat_carrier n sd I)
    hence id: "?rp = (int_mat_redpair :: nat \<Rightarrow> nat \<Rightarrow> ('f,int mat) lpoly_interL \<Rightarrow> ('f,'v)redpair_impl) n sd I" (is "_ = ?I I") by simp    
    have rp: "generic_redpair_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redpair_impl.generate_redpair[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id]]) 
  next
    case (Rat_mat_carrier d n sd I)
    hence id: "?rp = (rat_mat_redpair :: nat \<Rightarrow> nat \<Rightarrow> rat \<Rightarrow> ('f,rat mat) lpoly_interL \<Rightarrow> ('f,'v)redpair_impl) n sd d I" (is "_ = ?I I") by simp    
    have rp: "generic_redpair_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redpair_impl.generate_redpair[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id]]) 
  next
    case (Arc_mat_carrier n I)
    hence id: "?rp = (arctic_mat_redpair :: nat \<Rightarrow> ('f,arctic mat) lpoly_interL \<Rightarrow> ('f,'v)redpair_impl) n I" (is "_ = ?I I") by simp    
    have rp: "generic_redpair_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redpair_impl.generate_redpair[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id]]) 
  next
    case (Arc_rat_mat_carrier d n I)
    hence id: "?rp = (arctic_rat_mat_redpair :: nat \<Rightarrow> rat \<Rightarrow> ('f,arctic_rat mat) lpoly_interL \<Rightarrow> ('f,'v)redpair_impl) n d I" (is "_ = ?I I") by simp    
    have rp: "generic_redpair_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redpair_impl.generate_redpair[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id]]) 
  next
    case (LPO prec pi)
    let ?one = "(\<lambda> I. create_LPO_redpair lpo_repr_to_pr I) :: 'f lpo_repr \<Rightarrow> ('f filtered,'v)redpair_impl"
    have rp: "generic_redpair_impl ?one" ..
    let ?two = "(\<lambda> I. af_redpair pi (create_LPO_redpair lpo_repr_to_pr I)) :: 'f lpo_repr \<Rightarrow> ('f,'v)redpair_impl"
    from af_redpair[OF rp] 
    have rp: "generic_redpair_impl ?two" .
    from LPO have id: "?rp = ?two prec" by simp
    show ?thesis by(simp only: id, rule generic_redpair_impl.generate_redpair[of ?two, OF rp  v[simplified id] S[simplified id] NS[simplified id]])
  qed 
qed

end
    